function [ var ] = dqm( imSet, expT, hdrIm, tau, pyrweight )
% DQM Deghosting Quality Metric
%	This function calculates and returns the deghosting quality metric maps.
% 
% INPUTS:
% imSet     - 1-by-n cell of input RGB images (uint8 or uint16).
% expT      - 1-by-n vector of exposure times for input images.
% hdrIm     - Input RGB HDR image.
% tau       - Tau parameter (see paper).
% pyrweight - The weighting term for each level of the pyramid (see paper).
%
% OUTPUTS:
% var       - A struct of individual metric maps, has the following fieldnames:
%	var.bMap - Blending map (BM)
%	var.gradDirDiff - Gradient direction inconsistency map (GDM)
%	var.gradMagDiff - Gradient magnitude inconsistency map (GMM)
%	var.dynRegionDR - Dynamic regions' dynamic range (DR)
%
% To get individual metric scores compute pixelwise average of each artifact map.
% E.g. mean(var.bMap(:)) gives the score for the blending metric.
% For DR metric, higher score implies better quality. For all other metrics,
% lower score is better. In order to compute the unified metric you may take the
% weighted summation of individual metric scores using the weights given in
% Table 5 of our paper (in order to compute the UDQM correcty, you should
% multiply all metric scores with -1, except for DR before taking the weighted
% sum).
%
% SAMPLE RUN:
% imSet{1} = imread('sampleExposureSet/4.tiff');
% imSet{2} = imread('sampleExposureSet/5.tiff');
% imSet{3} = imread('sampleExposureSet/6.tiff');
% hdrIm = hdrread('sampleHDR/im.hdr');
% expT = [1/512 1/256 1/128];
% var = dqm(imSet, expT, hdrIm, 0.30, 4);
% figure, imshow(var.bMap);
% title('Blending Map (BM)');
% figure, imshow(var.gradDirDiff);
% title('Gradient Direction Inconsistency Map (GDM)');
% figure, imshow(var.gradMagDiff);
% title('Gradient Magnitude Inconsistency Map (GMM)');
% fprintf('DM: %g\n', var.dynRegionDR);

%   Version History:
%   1 Mar 2016    - Initial release.
%   13 April 2016 - Added an option to adjust the tolerance parameter of 
%					lsqnonneg. Removed unused globalDR variable. Fixed
%					typo in demo.m. 
%
%  Copyright (c) 2016, Okan Tarhan Tursun
%

%% Initialization
numIm = numel(imSet);
numBands = size(imSet{1},3);
middleExpIdx = floor((numIm-1) / 2) + 1;

%% Find well-exposedness maps
wExp = cell(numIm, 1);

for i = 1:numIm
    wExp{i} = bHat(im2double(imSet{i}));
    wExp{i} = min(wExp{i}, [], 3);
end
var.wExp = wExp;

%% Estimate the HDR reconstruction weights using least squares
radIm = cell(numIm, 1); % radiance maps (assuming linear CRF)

gamma = 1.0;
for i = 1:numIm
    radIm{i} = (im2double(imSet{i}).^gamma) ./ expT(i); % (2.^estExp(i)); % scaling between 0-1 and gamma correction
end

hdrIm = hdrIm ./ mean(hdrIm(:)) .* mean(radIm{middleExpIdx}(:)); % set the mean value of HDR to the mean value of middle exposed input image
if any(size(hdrIm) ~= size(imSet{1}))
    % workaround for methods using laplacian pyramid in the HDR
    % reconstruction step
    padsize = [size(imSet{1},1)-size(hdrIm,1) size(imSet{1},2)-size(hdrIm,2)];
    hdrIm = padarray(hdrIm, padsize, 0, 'post');
    for c =1:numBands
        hdrIm((end-padsize(1)):end, :, c) = ...
            radIm{middleExpIdx}((end-padsize(1)):end, :, c);
        hdrIm(:, (end-padsize(2)):end, c) = ...
            radIm{middleExpIdx}(:, (end-padsize(2)):end, c);
    end
end

%% HDR reconstruction weight estimation

radImCat = cat(3, radIm{:});
allIm = double(cat(3, hdrIm, radImCat));

bSize = [0 0]; % Border size [n m] implies (2*n+1)-by-(2*m+1) block size

warning('off', 'all');
options = optimset('lsqnonneg');
options.TolX = 1e-6;
weightMat = blockproc(allIm, [1 1], @(x) processor(x, numIm, numBands, options), ...
    'BorderSize', bSize, 'TrimBorder', false );
warning('on', 'all');

weightSparse = weightMat(:,:,1:end-1);

spams_rErr = weightMat(:,:,end);

weightSparse(weightSparse < 0) = 0;

sumWeight = sum(weightSparse,3);
normWeightSparse = weightSparse ./ repmat(sumWeight, [1 1 numIm]);
normWeightSparse(isnan(normWeightSparse)) = 0;
var.normWeightSparse = normWeightSparse;

%% Calculate blending-type artifact map
bMap = zeros([size(imSet{1}, 1) size(imSet{1},2) numIm*(numIm-1)/2]);
n = 0;
for i = 1:numIm-1
    for j = i+1:numIm
        n = n+1;
        radDiffRatio = h(radIm{i}, radIm{j});

        radDiffRatio(radDiffRatio <= tau) = 0;

        weightDiff = 1-abs(normWeightSparse(:,:,i) - normWeightSparse(:,:,j)); % function g
        avgWeight = mean(cat(3, normWeightSparse(:,:,i), normWeightSparse(:,:,j)), 3);

        bMap(:,:,n) = avgWeight .* weightDiff .* radDiffRatio .* wExp{i} .* wExp{j}; % blending type artifact map
    end
end
bMap = sum(bMap,3);
var.bMap = bMap;

%% Calculate the map of gradient magnitudes and directions
numlevels = 4;
gradMagDiff = 0;
gradDirDiff = 0;

maxGradVal = 0;
for i = 0:numlevels
    scale = 2.^(-1 * i);
    [gradDirDiffScale, gradMagDiffScale] = gradInconst(radIm, hdrIm, scale, tau);
    gradMagDiff = gradMagDiff + gradMagDiffScale * pyrweight.^i;
    gradDirDiff = gradDirDiff + gradDirDiffScale * pyrweight.^i;
    maxGradVal = maxGradVal + pyrweight.^i;
end

gradMagDiff = gradMagDiff ./ maxGradVal;
gradDirDiff = gradDirDiff ./ maxGradVal;

var.gradDirDiff = gradDirDiff;
var.gradMagDiff = gradMagDiff;

%% Calculate the dynamic range of the dynamic regions
mapDynRegions = zeros([size(imSet{1},1) size(imSet{1},2) numIm-1]);

for i=1:numIm-1
    mapDynRegions(:,:,i) = h(radIm{i}, radIm{i+1}) .* wExp{i} .* wExp{i+1};
end
mapDynRegions = max(mapDynRegions, [], 3);
var.mapDynRegions = mapDynRegions;
var.mapDynRegionsMask = mapDynRegions > tau;

hdrImDynRegions = hdrIm(repmat(mapDynRegions > tau, [1 1 3]));
prcRadDynRegions = prctile(hdrImDynRegions, [1 99]);
dynRegionDR = log10(prcRadDynRegions(2)) - log10(prcRadDynRegions(1));
var.dynRegionDR = dynRegionDR;

end

function result = h(E1, E2)

result = sum((E1-E2).^2,3).^0.5;
result = result ./ (sum(max(E1,E2).^2,3).^0.5 + 1E-6);

end

